import argparse
import subprocess
from carbontracker.tracker import CarbonTracker
import ast
import sys


def main():
    # Create a parser for the known arguments
    parser = argparse.ArgumentParser(description="CarbonTracker CLI", add_help=True)
    parser.add_argument("--log_dir", type=str, default="./logs", help="Log directory")
    parser.add_argument("--api_keys", type=str, help="API keys in a dictionary-like format, e.g., "
                                                 "'{\"electricitymaps\": \"YOUR_KEY\"}'", default=None)

    # Parse known arguments only
    known_args, remaining_args = parser.parse_known_args()

    # Parse the API keys string into a dictionary
    api_keys = ast.literal_eval(known_args.api_keys) if known_args.api_keys else None

    tracker = CarbonTracker(epochs=1, log_dir=known_args.log_dir, epochs_before_pred=0, api_keys=api_keys)
    tracker.epoch_start()

    # The remaining_args are considered as the command to execute
    if remaining_args:
        try:
            # Execute the command
            subprocess.run(remaining_args, check=True)
        except subprocess.CalledProcessError as e:
            print(f"Error executing command: {' '.join(remaining_args)}")
            print(f"Subprocess error: {e}")

    tracker.epoch_end()
    tracker.stop()


if __name__ == "__main__":
    main()




import numpy as np

from carbontracker import exceptions
from carbontracker.components.gpu import nvidia
from carbontracker.components.cpu import intel
from carbontracker.components.apple_silicon.powermetrics import AppleSiliconCPU, AppleSiliconGPU

COMPONENTS = [
    {
        "name": "gpu",
        "error": exceptions.GPUError("No GPU(s) available."),
        "handlers": [nvidia.NvidiaGPU, AppleSiliconGPU],
    },
    {
        "name": "cpu",
        "error": exceptions.CPUError("No CPU(s) available."),
        "handlers": [intel.IntelCPU, AppleSiliconCPU],
    },
]


def component_names():
    return [comp["name"] for comp in COMPONENTS]


def error_by_name(name):
    for comp in COMPONENTS:
        if comp["name"] == name:
            return comp["error"]


def handlers_by_name(name):
    for comp in COMPONENTS:
        if comp["name"] == name:
            return comp["handlers"]


class Component:
    def __init__(self, name, pids, devices_by_pid):
        self.name = name
        if name not in component_names():
            raise exceptions.ComponentNameError(f"No component found with name '{self.name}'.")
        self._handler = self._determine_handler(pids=pids, devices_by_pid=devices_by_pid)
        self.power_usages = []
        self.cur_epoch = -1  # Sentry

    @property
    def handler(self):
        if self._handler is None:
            raise error_by_name(self.name)
        return self._handler

    def _determine_handler(self, pids, devices_by_pid):
        handlers = handlers_by_name(self.name)
        for h in handlers:
            handler = h(pids=pids, devices_by_pid=devices_by_pid)
            if handler.available():
                return handler
        return None

    def devices(self):
        return self.handler.devices()

    def available(self):
        return self._handler is not None

    def collect_power_usage(self, epoch):
        if epoch < 1:
            return

        if epoch != self.cur_epoch:
            self.cur_epoch = epoch
            # If we haven't measured for some epochs due to too slow
            # update_interval, we copy previous epoch measurements s.t.
            # there exists measurements for every epoch.
            diff = self.cur_epoch - len(self.power_usages) - 1
            if diff != 0:
                for _ in range(diff):
                    # Copy previous measurement lists.
                    latest_measurements = self.power_usages[-1] if self.power_usages else []
                    self.power_usages.append(latest_measurements)
            self.power_usages.append([])
        try:
            self.power_usages[-1].append(self.handler.power_usage())
        except exceptions.IntelRaplPermissionError:
            # Only raise error if no measurements have been collected.
            if not self.power_usages[-1]:
                print(
                    "No sudo access to read Intel's RAPL measurements from the energy_uj file."
                    "\nSee issue: https://github.com/lfwa/carbontracker/issues/40"
                )
            # Append zero measurement to avoid further errors.
            self.power_usages.append([0])
        except exceptions.GPUPowerUsageRetrievalError:
            if not self.power_usages[-1]:
                print(
                    "GPU model does not support retrieval of power usages in NVML."
                    "\nSee issue: https://github.com/lfwa/carbontracker/issues/36"
                )
                # Append zero measurement to avoid further errors.
                self.power_usages.append([0])

    def energy_usage(self, epoch_times):
        """Returns energy (mWh) used by component per epoch."""
        energy_usages = []
        # We have to compute each epoch in a for loop since numpy cannot
        # handle lists of uneven length.
        for idx, (power, time) in enumerate(zip(self.power_usages, epoch_times)):
            # If no power measurement exists, try to use measurements from
            # later epochs.
            while not power and idx != len(self.power_usages) - 1:
                idx += 1
                power = self.power_usages[idx]
            if not power:
                power = [[0]]
            avg_power_usage = np.mean(power, axis=0)
            energy_usage = np.multiply(avg_power_usage, time).sum()
            # Convert from J to kWh.
            if energy_usage != 0:
                energy_usage /= 3600000
            energy_usages.append(energy_usage)

        # Ensure energy_usages and epoch_times have same length by
        # copying latest measurement if it exists.
        diff = len(epoch_times) - len(energy_usages)
        if diff != 0:
            for _ in range(0, diff):
                # TODO: Warn that no measurements have been fetched.
                latest_energy = energy_usages[-1] if energy_usages else 0
                energy_usages.append(latest_energy)

        return energy_usages

    def init(self):
        self.handler.init()

    def shutdown(self):
        self.handler.shutdown()


def create_components(components, pids, devices_by_pid):
    components = components.strip().replace(" ", "").lower()
    if components == "all":
        return [Component(name=comp_name, pids=pids, devices_by_pid=devices_by_pid) for comp_name in component_names()]
    else:
        return [
            Component(name=comp_name, pids=pids, devices_by_pid=devices_by_pid) for comp_name in components.split(",")
        ]



# Power Usage Effectiveness Coefficient.
# Rhonda Ascierto. 2020-2021. Uptime Institute Annual Data Center Survey.
# https://www.missioncriticalmagazine.com/ext/resources/whitepapers/2020/2020AnnualSurvey_EndUser_v4s.pdf
PUE_2020 = 1.59

# https://uptimeinstitute.com/2021-data-center-industry-survey-results
PUE_2021 = 1.57

# https://uptimeinstitute.com/uptime_assets/6768eca6a75d792c8eeede827d76de0d0380dee6b5ced20fde45787dd3688bfe-2022-data-center-industry-survey-en.pdf
PUE_2022 = 1.55

# World-wide average carbon intensity of electricity production in 2019.
# https://www.iea.org/reports/global-energy-co2-status-report-2019/emissions
WORLD_2019_CARBON_INTENSITY = 475



import platform
import subprocess
import re
import time
from carbontracker.components.handler import Handler


class PowerMetricsUnified:
    _output = None
    _last_updated = None

    @staticmethod
    def get_output():
        if PowerMetricsUnified._output is None or time.time() - PowerMetricsUnified._last_updated > 1:
            PowerMetricsUnified._output = subprocess.check_output(
                ["sudo", "powermetrics", "-n", "1", "-i", "1000", "--samplers", "all"], universal_newlines=True
            )
            PowerMetricsUnified._last_updated = time.time()
        return PowerMetricsUnified._output


class AppleSiliconCPU(Handler):
    def init(self, pids=None, devices_by_pid=None):
        self.devices_list = ["CPU"]
        self.cpu_pattern = re.compile(r"CPU Power: (\d+) mW")

    def shutdown(self):
        pass

    def devices(self):
        """Returns a list of devices (str) associated with the component."""
        return self.devices_list

    def available(self):
        return platform.system() == "Darwin"

    def power_usage(self):
        output = PowerMetricsUnified.get_output()
        cpu_power = self.parse_power(output, self.cpu_pattern)
        return cpu_power

    def parse_power(self, output, pattern):
        match = pattern.search(output)
        if match:
            power = float(match.group(1)) / 1000  # Convert mW to W
            return power
        else:
            return 0.0


class AppleSiliconGPU(Handler):
    def init(self, pids=None, devices_by_pid=None):
        self.devices_list = ["GPU", "ANE"]
        self.gpu_pattern = re.compile(r"GPU Power: (\d+) mW")
        self.ane_pattern = re.compile(r"ANE Power: (\d+) mW")

    def devices(self):
        """Returns a list of devices (str) associated with the component."""
        return self.devices_list

    def available(self):
        return platform.system() == "Darwin"

    def power_usage(self):
        output = PowerMetricsUnified.get_output()
        gpu_power = self.parse_power(output, self.gpu_pattern)
        ane_power = self.parse_power(output, self.ane_pattern)
        return gpu_power + ane_power

    def parse_power(self, output, pattern):
        match = pattern.search(output)
        if match:
            power = float(match.group(1)) / 1000  # Convert mW to W (J/s)
            return power
        else:
            return 0.0



"""Utilities to query NVIDIA GPUs.

This module provides utilities to query NVIDIA GPUs using NVIDIA Management
Library (NVML). It is important to run nvmlInit() before any queries are made
and nvmlShutdown() after all queries are finished. For performance, it is
recommended to run nvmlInit() and nvmlShutdown() as few times as possible, e.g.
by running queries in batches (initializing and shutdown after each query can
result in more than a 10x slowdown).
"""
import sys

import pynvml
import os

from carbontracker import exceptions
from carbontracker.components.handler import Handler


class NvidiaGPU(Handler):
    def devices(self):
        """
        Note:
            Requires NVML to be initialized.
        """
        names = [pynvml.nvmlDeviceGetName(handle) for handle in self._handles]

        # Decode names if Python version is less than 3.9
        if sys.version_info < (3, 9):
            names = [name.decode("utf-8") for name in names]

        return names

    def available(self):
        """Checks if NVML and any GPUs are available."""
        try:
            self.init()
            if len(self._handles) > 0:
                available = True
            else:
                available = False
            self.shutdown()
        except pynvml.NVMLError:
            available = False
        return available

    def power_usage(self):
        """Retrieves instantaneous power usages (W) of all GPUs in a list.

        Note:
            Requires NVML to be initialized.
        """
        gpu_power_usages = []

        for handle in self._handles:
            try:
                # Retrieves power usage in mW, divide by 1000 to get in W.
                power_usage = pynvml.nvmlDeviceGetPowerUsage(handle) / 1000
                gpu_power_usages.append(power_usage)
            except pynvml.NVMLError:
                raise exceptions.GPUPowerUsageRetrievalError()
        return gpu_power_usages

    def init(self):
        pynvml.nvmlInit()
        if self.devices_by_pid:
            self._handles = self._get_handles_by_pid()
        else:
            self._handles = self._get_handles()

    def shutdown(self):
        pynvml.nvmlShutdown()

    def _get_handles(self):
        """Returns handles of GPUs in slurm job if existent otherwise all
        available GPUs."""
        device_indices = self._slurm_gpu_indices()

        # If we cannot retrieve indices from slurm then we retrieve all GPUs.
        if not device_indices:
            device_count = pynvml.nvmlDeviceGetCount()
            device_indices = range(device_count)

        return [pynvml.nvmlDeviceGetHandleByIndex(i) for i in device_indices]

    def _slurm_gpu_indices(self):
        """Returns indices of GPUs for the current slurm job if existent.

        Note:
            Relies on the environment variable CUDA_VISIBLE_DEVICES to not
            overwritten. Alternative variables could be SLURM_JOB_GPUS and
            GPU_DEVICE_ORDINAL.
        """
        index_str = os.environ.get("CUDA_VISIBLE_DEVICES")
        try:
            indices = [int(i) for i in index_str.split(",")]
        except:
            indices = None
        return indices

    def _get_handles_by_pid(self):
        """Returns handles of GPU running at least one process from PIDS.

        Note:
            GPUs need to have started work before showing any processes.
            Requires NVML to be initialized.
            Bug: Containers need to be started with --pid=host for NVML to show
            processes: https://github.com/NVIDIA/nvidia-docker/issues/179.
        """
        device_count = pynvml.nvmlDeviceGetCount()
        devices = []

        for index in range(device_count):
            handle = pynvml.nvmlDeviceGetHandleByIndex(index)
            gpu_pids = [
                p.pid
                for p in pynvml.nvmlDeviceGetComputeRunningProcesses(handle)
                + pynvml.nvmlDeviceGetGraphicsRunningProcesses(handle)
            ]

            if set(gpu_pids).intersection(self.pids):
                devices.append(handle)

        return devices



from abc import ABCMeta, abstractmethod


class Handler:
    __metaclass__ = ABCMeta

    def __init__(self, pids, devices_by_pid):
        self.pids = pids
        self.devices_by_pid = devices_by_pid

    @abstractmethod
    def devices(self):
        """Returns a list of devices (str) associated with the component."""
        raise NotImplementedError

    @abstractmethod
    def available(self):
        """Returns True if the handler is available."""
        raise NotImplementedError

    @abstractmethod
    def power_usage(self):
        """Returns the current power usage (W) in a list."""
        raise NotImplementedError

    @abstractmethod
    def init(self):
        """Initializes the handler."""
        raise NotImplementedError

    @abstractmethod
    def shutdown(self):
        """Shuts down the handler."""
        raise NotImplementedError



"""Script to create a short csv file with carbon intensity data.

Uses csv from https://ourworldindata.org/grapher/carbon-intensity-electricity.
"""
import argparse

import pycountry
import pandas as pd


def country_to_alpha2(name):
    try:
        return pycountry.countries.get(name=name).alpha_2
    except Exception as err:
        return None


def main(args):
    intensity_df = pd.read_csv(args.input_csv)
    intensity_df.sort_values(by="Year", inplace=True, ascending=False)

    # Add ISO 3166-1 alpha-2 country codes.
    intensity_df["alpha-2"] = intensity_df["Entity"].apply(country_to_alpha2)

    # Drop rows with None or NaN.
    intensity_df.dropna(subset=["alpha-2"], inplace=True)

    # Only keep most recent year from every country.
    # Note that no country shares alpha-2 code, so should not lose any info.
    intensity_df = intensity_df.groupby("alpha-2").first()

    # Save to csv.
    intensity_df.to_csv(args.output_csv)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--input_csv", help="Path to input csv.", required=True)
    parser.add_argument("--output_csv", help="Path to output csv.", required=True)
    args = parser.parse_args()
    main(args)



import logging
import os
import sys
import pathlib
import datetime
import importlib_metadata as metadata
from carbontracker import constants


def convert_to_timestring(seconds, add_milliseconds=False):
    negative = False
    if seconds < 0:
        negative = True
        seconds = abs(seconds)

    m, s = divmod(seconds, 60)
    if not add_milliseconds:
        s = int(round(s))
        if s == 60:
            m += 1
            s = 0
    else:
        if f"{s:05.2f}"[0:2] == "60":
            m += 1
            s = 0
    h, m = divmod(m, 60)
    h = int(h)
    m = int(m)
    if not add_milliseconds:
        return f"-{h:d}:{m:02d}:{s:02d}" if negative else f"{h:d}:{m:02d}:{s:02d}"
    else:
        return f"-{h:d}:{m:02d}:{s:05.2f}" if negative else f"{h:d}:{m:02d}:{s:05.2f}"


class TrackerFormatter(logging.Formatter):
    converter = datetime.datetime.fromtimestamp

    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        if datefmt:
            s = ct.strftime(datefmt)
        else:
            t = ct.strftime("%Y-%m-%d %H:%M:%S")
            s = "%s" % t
        return s


class VerboseFilter(logging.Filter):
    def __init__(self, verbose):
        super().__init__()
        self.verbose = verbose

    def filter(self, record):
        return self.verbose > 0


class Logger:
    def __init__(self, log_dir=None, verbose=0, log_prefix=""):
        self.verbose = verbose
        self.logger, self.logger_output, self.logger_err = self._setup(log_dir=log_dir, log_prefix=log_prefix)
        self._log_initial_info()
        self.msg_prepend = "CarbonTracker: "

    def _setup(self, log_dir=None, log_prefix=""):
        if log_prefix:
            log_prefix += "_"

        logger_name = f"{log_prefix}{os.getpid()}"
        logger = logging.getLogger(logger_name)

        logger_err = logging.getLogger("carbontracker.err")
        logger_output = logging.getLogger("carbontracker.output")
        logger.propagate = False
        logger.setLevel(logging.DEBUG)
        logger_output.propagate = False
        logger_output.setLevel(logging.DEBUG)
        logger_err.propagate = False
        logger_err.setLevel(logging.DEBUG)

        ch = logging.StreamHandler(stream=sys.stdout)
        c_formatter = logging.Formatter("{message}", style="{")
        ch.setLevel(logging.INFO)
        ch.setFormatter(c_formatter)
        ch.addFilter(VerboseFilter(self.verbose))
        logger_output.addHandler(ch)

        # Add error logging to console.
        ce = logging.StreamHandler(stream=sys.stdout)
        ce_formatter = logging.Formatter("CarbonTracker: {levelname} - {message}", style="{")
        ce.setLevel(logging.INFO)
        ce.setFormatter(ce_formatter)
        logger_err.addHandler(ce)

        if log_dir is not None:
            # Create logging directory if it does not exist.
            pathlib.Path(log_dir).mkdir(parents=True, exist_ok=True)

            # Modified ISO8601 format YYYY-MM-DDThhmmssZ.
            # : does not work in file naming on Windows.
            date_format = "%Y-%m-%dT%H%M%SZ"
            date = datetime.datetime.now().strftime(date_format)

            f_formatter = TrackerFormatter(fmt="%(asctime)s - %(message)s")

            # Add output logging to file.
            fh = logging.FileHandler(f"{log_dir}/{logger_name}_{date}_carbontracker_output.log")
            fh.setLevel(logging.INFO)
            fh.setFormatter(f_formatter)
            logger_output.addHandler(fh)

            # Add standard logging to file.
            f = logging.FileHandler(f"{log_dir}/{logger_name}_{date}_carbontracker.log")
            f.setLevel(logging.DEBUG)
            f.setFormatter(f_formatter)
            logger.addHandler(f)

            # Add error logging to file.
            err_formatter = logging.Formatter("{asctime} - {threadName} - {levelname} - {message}", style="{")
            f_err = logging.FileHandler(f"{log_dir}/{logger_name}_{date}_carbontracker_err.log", delay=True)
            f_err.setLevel(logging.DEBUG)
            f_err.setFormatter(err_formatter)
            logger_err.addHandler(f_err)

        return logger, logger_output, logger_err

    def _log_initial_info(self):
        self.info(f"{__package__} version {metadata.version(__package__)}")
        self.info(
            "Only predicted and actual consumptions are multiplied by a PUE "
            f"coefficient of {constants.PUE_2022} (Rhonda Ascierto, 2022, Uptime "
            "Institute Global Data Center Survey)."
        )

    def output(self, msg, verbose_level=0):
        self.logger_output.info(self.msg_prepend + msg)

    def info(self, msg):
        self.logger.info(msg)

    def err_debug(self, msg):
        self.logger_err.debug(msg)

    def err_info(self, msg):
        self.logger_err.info(msg)

    def err_warn(self, msg):
        self.logger_err.warn(msg)

    def err_critical(self, msg):
        self.logger_err.critical(msg)



import requests

from carbontracker import exceptions
from carbontracker.emissions.intensity.fetcher import IntensityFetcher
from carbontracker.emissions.intensity import intensity

API_URL = "https://api-access.electricitymaps.com/free-tier/carbon-intensity/latest"


class ElectricityMap(IntensityFetcher):
    _api_key = None

    @classmethod
    def set_api_key(cls, key):
        cls._api_key = key

    def suitable(self, g_location):
        return self._api_key is not None

    def carbon_intensity(self, g_location, time_dur=None):
        carbon_intensity = intensity.CarbonIntensity(g_location=g_location)

        try:
            ci = self._carbon_intensity_by_location(lon=g_location.lng, lat=g_location.lat)
        except:
            ci = self._carbon_intensity_by_location(zone=g_location.country)

        carbon_intensity.carbon_intensity = ci

        return carbon_intensity

    def _carbon_intensity_by_location(self, lon=None, lat=None, zone=None):
        """Retrieves carbon intensity (gCO2eq/kWh) by location.

        Note:
            Only use arguments (lon, lat) or country_code.

        Args:
            lon (float): Longitude. Defaults to None.
            lat (float): Lattitude. Defaults to None.
            zone (str): Alpha-2 country code. Defaults to None.

        Returns:
            Carbon intensity in gCO2eq/kWh.
        """
        if zone is not None:
            params = (("zone", zone),)
            assert lon is None and lat is None
        elif lon is not None and lat is not None:
            params = (("lon", lon), ("lat", lat))
            assert zone is None

        headers = {"auth-token": self._api_key}

        response = requests.get(API_URL, headers=headers, params=params)
        if not response.ok:
            raise exceptions.CarbonIntensityFetcherError(response.json())
        carbon_intensity = response.json()["carbonIntensity"]

        return carbon_intensity



import numpy as np


# TODO: Do advanced prediction based on profiling work.
def predict_energy(total_epochs, epoch_energy_usages):
    avg_epoch_energy = np.mean(epoch_energy_usages)
    return total_epochs * avg_epoch_energy


def predict_time(total_epochs, epoch_times):
    avg_epoch_time = np.mean(epoch_times)
    return total_epochs * avg_epoch_time



import traceback

import geocoder
import pkg_resources
import numpy as np
import pandas as pd

from carbontracker import loggerutil
from carbontracker import exceptions
from carbontracker import constants
from carbontracker.emissions.intensity.fetchers import carbonintensitygb
from carbontracker.emissions.intensity.fetchers import energidataservice
from carbontracker.emissions.intensity.fetchers import electricitymaps


def get_default_intensity():
    """Retrieve static default carbon intensity value based on location."""
    try:
        g_location = geocoder.ip("me")
        if not g_location.ok:
            raise exceptions.IPLocationError("Failed to retrieve location based on IP.")
        address = g_location.address
        country = g_location.country
    except Exception as err:
        address = "Unknown"
        country = "Unknown"

    try:
        carbon_intensities_df = pd.read_csv(
            pkg_resources.resource_filename("carbontracker", "data/carbon-intensities.csv")
        )
        intensity_row = carbon_intensities_df[carbon_intensities_df["alpha-2"] == country].iloc[0]
        intensity = intensity_row["Carbon intensity of electricity (gCO2/kWh)"]
        year = intensity_row["Year"]
        description = f"Defaulted to average carbon intensity for {country} in {year} of {intensity:.2f} gCO2/kWh."
    except Exception as err:
        intensity = constants.WORLD_2019_CARBON_INTENSITY
        description = f"Defaulted to average carbon intensity for world in 2019 of {intensity:.2f} gCO2/kWh."

    description = f"Live carbon intensity could not be fetched at detected location: {address}. " + description
    default_intensity = {
        "carbon_intensity": intensity,
        "description": description,
    }
    return default_intensity


default_intensity = get_default_intensity()


class CarbonIntensity:
    def __init__(
        self,
        carbon_intensity=None,
        g_location=None,
        address="UNDETECTED",
        message=None,
        success=False,
        is_prediction=False,
        default=False,
    ):
        self.carbon_intensity = carbon_intensity
        self.g_location = g_location
        self.address = address
        self.message = message
        self.success = success
        self.is_prediction = is_prediction
        if default:
            self.set_as_default()

    def set_as_default(self):
        self.set_default_intensity()
        self.set_default_message()

    def set_default_intensity(self):
        self.carbon_intensity = default_intensity["carbon_intensity"]

    def set_default_message(self):
        self.message = default_intensity["description"]


def carbon_intensity(logger, time_dur=None):
    fetchers = [
        electricitymaps.ElectricityMap(),
        energidataservice.EnergiDataService(),
        carbonintensitygb.CarbonIntensityGB(),
    ]

    carbon_intensity = CarbonIntensity(default=True)

    try:
        g_location = geocoder.ip("me")
        if not g_location.ok:
            raise exceptions.IPLocationError("Failed to retrieve location based on IP.")
        carbon_intensity.address = g_location.address
    except:
        err_str = traceback.format_exc()
        logger.err_info(err_str)
        return carbon_intensity

    for fetcher in fetchers:
        if not fetcher.suitable(g_location):
            continue
        try:
            carbon_intensity = fetcher.carbon_intensity(g_location, time_dur)
            if not np.isnan(carbon_intensity.carbon_intensity):
                carbon_intensity.success = True
                set_carbon_intensity_message(carbon_intensity, time_dur)
            carbon_intensity.address = g_location.address
        except:
            err_str = traceback.format_exc()
            logger.err_info(err_str)

    if not carbon_intensity.success:
        logger.err_warn(
            "Failed to retrieve carbon intensity: Defaulting to average carbon intensity {} gCO2/kWh.".format(
                default_intensity["carbon_intensity"]
            )
        )
    return carbon_intensity


def set_carbon_intensity_message(ci, time_dur):
    if ci.is_prediction:
        if ci.success:
            ci.message = (
                "Carbon intensity for the next "
                f"{loggerutil.convert_to_timestring(time_dur)} is "
                f"predicted to be {ci.carbon_intensity:.2f} gCO2/kWh"
            )
        else:
            ci.message = (
                "Failed to predict carbon intensity for the next "
                f"{loggerutil.convert_to_timestring(time_dur)}, "
                f"fallback on average measured intensity"
            )
    else:
        if ci.success:
            ci.message = f"Current carbon intensity is {ci.carbon_intensity:.2f} gCO2/kWh"
        else:
            ci.set_default_message()
    ci.message += f" at detected location: {ci.address}."



import pandas as pd
import os

here = os.path.abspath(os.path.dirname(__file__))
conversion_file = os.path.join(here, "co2eq.csv")
CONVERSION_DF = pd.read_csv(conversion_file)


def convert(g_co2eq):
    """Converts gCO2eq to all units in range specified by CONVERSION_FILE."""
    conversions = []
    df = CONVERSION_DF.iloc[-1:]  # Use latest conversion factors
    converters = df.loc[(df["lowerbound"] <= g_co2eq) & (df["upperbound"] >= g_co2eq)]
    for _, converter in converters.iterrows():
        units = g_co2eq / converter["gCO2eq_per_unit"]
        conversions.append((units, converter["unit"]))
    return conversions
